import io
import logging
from threading import Lock
from typing import ClassVar

from flask import send_file
from mxcubecore import HardwareRepository as HWR
from mxcubecore import queue_entry as qe
from mxcubecore.HardwareObjects.abstract.AbstractSampleView import AbstractSampleView
from mxcubecore.queue_entry.base_queue_entry import CENTRING_METHOD

from mxcubeweb.core.adapter.adapter_base import AdapterBase
from mxcubeweb.core.models.adaptermodels import Base64StrModel, ListOfShapesModel
from mxcubeweb.core.models.configmodels import ResourceHandlerConfigModel
from mxcubeweb.core.util.convertutils import (
    from_camel,
    to_camel,
)

logger = logging.getLogger("MX3.HWR")

resource_handler_config = ResourceHandlerConfigModel(
    name="sample_view",
    attributes=[
        "get_value",
        "sample_image_meta_data",
        "shapes",
    ],
    commands=[
        "snapshot",
        "set_image_size",
        "move_to_centred_position",
        "update_shapes",
        "delete_shape",
        "rotate_to",
        "centre_click",
        "abort_centring",
        "click",
        "accept_centring",
        "move_to_beam",
        "set_centring_method",
        "start_click_centring",
    ],
)


class SampleViewAdapter(AdapterBase):
    """Adapter for AbstractSampleView — routes moved into adapter methods."""

    SUPPORTED_TYPES: ClassVar[list[object]] = [AbstractSampleView]

    def __init__(self, ho, role, app):
        super().__init__(ho, role, app, resource_handler_config)
        self._click_count = 0
        self._click_limit = int(HWR.beamline.config.click_centring_num_clicks or 3)
        self._centring_point_id = None
        self._error = False
        self._click_lock = Lock()
        self._click_in_progress = False

        self._ho.connect("shapesChanged", self._emit_shapes_updated)
        self._ho.connect("newGridResult", self._handle_grid_result)

        self._ho.connect("centringStarted", self._centring_started)
        self._ho.connect("centringSuccessful", self._wait_for_centring_finishes)
        self._ho.connect("centringFailed", self._wait_for_centring_finishes)
        self._ho.connect("centringAccepted", self._centring_add_current_point)

        zoom_motor = HWR.beamline.diffractometer.zoom

        if zoom_motor:
            zoom_motor.connect("stateChanged", self._zoom_changed)

        HWR.beamline.queue_manager.connect(
            "queue_execution_finished",
            self._queue_execution_finished_handler,
        )

    def _queue_execution_finished_handler(self, queue_state=None):  # noqa: ARG002
        self._emit_shapes_updated()

    def _centring_add_current_point(self, *args):
        shape = self._ho.get_shape(self._centring_point_id)

        # There is no current centered point shape when the centring is done
        # by software like Workflows, so we add one.
        if not shape:
            if len(args) > 1 and isinstance(args[1], dict):
                motor_positions = args[1].get("motors")
            else:
                motor_positions = self._ho.get_positions()
            (x, y) = self._ho.motor_positions_to_screen(motor_positions)
            self._centring_update_current_point(motor_positions, x, y)
            shape = self._ho.get_shape(self._centring_point_id)

        if shape:
            shape.state = "SAVED"
            self._emit_shapes_updated()
            self._centring_point_id = None
        else:
            logging.getLogger("MX3.HWR").exception("Centring failed !")

    def _centring_update_current_point(self, motor_positions, x, y):
        point = self._ho.get_shape(self._centring_point_id)

        if point:
            point.move_to_mpos([motor_positions], [x, y])
        else:
            point = self._ho.add_shape_from_mpos([motor_positions], (x, y), "P")
            point.state = "TMP"
            point.selected = True
            self._centring_point_id = point.id

        self._emit_shapes_updated()

    def centring_remove_current_point(self):
        if self._centring_point_id:
            HWR.beamline.sample_view.delete_shape(self._centring_point_id)
            self._emit_shapes_updated()
            self._centring_point_id = None

    def _wait_for_centring_finishes(self, *args, **kwargs):  # noqa: ARG002
        """Executed when a centring is finished."""
        try:
            centring_status = args[1]
        except IndexError:
            centring_status = {"valid": False}

        # we do not send/save any centring data if there is no sample
        # to avoid the 2d centring when no sample is mounted
        if self.app.lims.get_current_sample().get("sampleID", "") == "":
            return

        # If centering is valid add the point, otherwise remove it
        if centring_status["valid"]:
            motor_positions = centring_status["motors"]
            motor_positions.pop("zoom", None)
            motor_positions.pop("beam_y", None)
            motor_positions.pop("beam_x", None)

            (x, y) = self._ho.motor_positions_to_screen(motor_positions)

            self._centring_update_current_point(motor_positions, x, y)

            if self.app.AUTO_MOUNT_SAMPLE:
                self._ho.accept_centring()
        else:
            self._error = True

    def _centring_started(self, method, *args):  # noqa: ARG002
        msg = {"method": method}

        if method in ["Manual"]:
            msg = {"method": qe.CENTRING_METHOD.MANUAL}
        else:
            msg = {"method": qe.CENTRING_METHOD.LOOP}

        self.app.server.emit("sample_centring", msg, namespace="/hwr")

    def _zoom_changed(self, *args, **kwargs):  # noqa: ARG002
        ppm = HWR.beamline.diffractometer.get_pixels_per_mm()
        self.app.server.emit(
            "update_pixels_per_mm",
            {"pixelsPerMm": ppm},
            namespace="/hwr",
        )

    def _emit_shapes_updated(self):
        shape_dict = {}

        for shape in self._ho.get_shapes():
            _s = to_camel(shape.as_dict())
            shape_dict.update({shape.id: _s})

        self.app.server.emit("update_shapes", {"shapes": shape_dict}, namespace="/hwr")

    def _handle_grid_result(self, shape):
        self.app.server.emit(
            "grid_result_available", {"shape": shape}, namespace="/hwr"
        )

    def centring_clicks_left(self):
        return self._click_limit - self._click_count

    def centring_reset_click_count(self):
        self._click_count = 0

    def centring_click(self):
        self._click_count += 1

    def get_value(self) -> dict:
        return {}

    def sample_image_meta_data(self) -> dict:
        """Get information about current sample image.

        Returns a dictionary with the format:

            data = {"pixelsPerMm": pixelsPerMm,
                    "imageWidth": width,
                    "imageHeight": height,
                    "format": fmt,
                    "sourceIsScalable": source_is_scalable,
                    "scale": scale,
                    "videoSizes": video_sizes,
                    "position": position,
                    "shape": shape,
                    "size_x": sx, "size_y": sy}

        :returns: Dictionary with view port data, with format described above
        :rtype: dict
        """
        fmt, source_is_scalable = "MJPEG", False

        if self.app.CONFIG.app.VIDEO_FORMAT == "MPEG1":
            fmt, source_is_scalable = "MPEG1", True
            video_sizes = self._ho.camera.get_available_stream_sizes()
            (width, height, scale) = self._ho.camera.get_stream_size()
        else:
            scale = 1
            width = self._ho.camera.get_width()
            height = self._ho.camera.get_height()
            video_sizes = [(width, height)]

        pixels_per_mm = HWR.beamline.diffractometer.get_pixels_per_mm()

        return {
            "pixelsPerMm": pixels_per_mm,
            "imageWidth": width,
            "imageHeight": height,
            "format": fmt,
            "sourceIsScalable": source_is_scalable,
            "scale": scale,
            "videoSizes": video_sizes,
            "videoHash": self._ho.camera.stream_hash,
            "videoURL": self.app.CONFIG.app.VIDEO_STREAM_URL,
        }

    def shapes(self) -> list:
        return {shape.id: to_camel(shape.as_dict()) for shape in self._ho.get_shapes()}

    def snapshot(self, overlay: Base64StrModel):
        """Take snapshot of the sample view.

        data = {"overlay": overlay_data} overlay is the image data to overlay on sample
        image, it should normally contain the data of shapes drawn on canvas.
        Return: Overlayed image uri, if successful, statuscode 500 otherwise.
        """
        mimetype, overlay_data = overlay.value.split(",")

        if "image/jpeg" not in mimetype:
            msg = "Image type should be jpeg"
            self.app.server_logger.error("Taking a snapshot failed")
            self.app.server_logger.error(msg)
            raise ValueError(msg)

        try:
            image = self._ho.take_snapshot(overlay_data=overlay_data)
            data = io.BytesIO()
            image.save(data, "JPEG")
            data.seek(0)

            return send_file(
                data,
                mimetype="image/jpeg",
                as_attachment=True,
                download_name="snapshot.jpeg",
            )
        except Exception as e:
            self.app.server_logger.exception("Taking a snapshot failed")
            msg = "Taking a snapshot failed"
            raise RuntimeError(msg) from e

    def set_image_size(self, width: float, height: float) -> None:
        HWR.beamline.sample_view.camera.restart_streaming((width, height))
        return self.sample_image_meta_data()

    def move_to_centred_position(self, point_id: str):
        point = self._ho.get_shape(point_id)

        if point:
            motor_positions = point.get_centred_position().as_dict()
            HWR.beamline.diffractometer.set_value_motors(motor_positions)
            HWR.beamline.diffractometer.save_centring_positions()
            return {}
        msg = "Could not move to centred position"
        raise RuntimeError(msg)

    def update_shapes(self, shapes: ListOfShapesModel) -> dict:
        updated_shapes = []
        for s in shapes.shapes:
            shape_data = from_camel(s)
            pos = []

            # Get the shape if already exists
            shape = self._ho.get_shape(shape_data.get("id", -1))

            # If shape does not exist add it
            if not shape:
                refs, t = shape_data.pop("refs", []), shape_data.pop("t", "")
                state = shape_data.pop("state", "SAVED")
                user_state = shape_data.pop("user_state", "SAVED")

                # Store pixels per mm for third party software, to facilitate
                # certain calculations

                shape_data["pixels_per_mm"] = (
                    HWR.beamline.diffractometer.get_pixels_per_mm()
                )

                shape_data["beam_pos"] = (
                    HWR.beamline.beam.get_beam_position_on_screen()[0],
                    HWR.beamline.beam.get_beam_position_on_screen()[1],
                )
                shape_data["beam_width"] = HWR.beamline.beam.get_value()[0]
                shape_data["beam_height"] = HWR.beamline.beam.get_value()[1]

                # Shape does not have any refs, create a new Centered position
                if not refs:
                    try:
                        x, y = shape_data["screen_coord"]
                        mpos = self._ho.get_centred_point_from_coord(
                            x, y, return_by_names=True
                        )
                        pos.append(mpos)

                        # We also store the center of the grid
                        if t == "G":
                            # coords for the center of the grid
                            x_c = (
                                x
                                + (shape_data["num_cols"] / 2.0)
                                * shape_data["cell_width"]
                            )
                            y_c = (
                                y
                                + (shape_data["num_rows"] / 2.0)
                                * shape_data["cell_height"]
                            )
                            center_positions = self._ho.get_centred_point_from_coord(
                                x_c, y_c, return_by_names=True
                            )
                            pos.append(center_positions)

                        shape = self._ho.add_shape_from_mpos(
                            pos, (x, y), t, state, user_state
                        )
                    except Exception:
                        msg = "Could not create shape from screen coordinates"
                        logging.getLogger("MX3.HWR").exception(msg)
                        logging.getLogger("MX3.HWR").exception("")

                else:
                    shape = self._ho.add_shape_from_refs(refs, t, state, user_state)

            # shape will be none if creation failed, so we check if shape exists
            # before setting additional parameters
            if shape:
                shape.update_from_dict(shape_data)
                shape_dict = to_camel(shape.as_dict())
                updated_shapes.append(shape_dict)

        return {"shapes": updated_shapes}

    def delete_shape(self, sid: str):
        self._ho.delete_shape(sid)
        return {}

    def rotate_to(self, sid: str):
        if sid:
            shape = self._ho.get_shape(sid)
            cp = shape.get_centred_position()
            omega_value = round(float(cp.as_dict().get("omega", None)), 3)

            if omega_value:
                try:
                    HWR.beamline.diffractometer.omega.set_value(omega_value)
                except Exception as e:
                    msg = "Rotate to shape failed"
                    raise RuntimeError(msg) from e

        return {}

    def centre_click(self):
        try:
            return self._ho.start_manual_centring()
        except Exception as e:
            msg = "Could not start %s click centring"
            self.app.server_logger.exception(
                msg, self.app.beamline.config.click_centring_num_clicks
            )
            msg = "Could not start click centring"
            raise RuntimeError(msg) from e

    def abort_centring(self):
        try:
            logging.getLogger("user_level_log").info("User canceled centring")
            self._ho.cancel_centring()
            self.centring_remove_current_point()
            self._click_in_progress = False
        except Exception:  # noqa: BLE001
            logging.getLogger("MX3.HWR").warning("Canceling centring failed")

        return {}

    def click(self, x: float, y: float):
        with self._click_lock:
            if self._click_in_progress:
                return {"clicksLeft": self.centring_clicks_left()}

            self._click_in_progress = True

        if self._ho.current_centring_procedure:
            try:
                self.centring_click()
                self._ho.image_clicked(x, y)
                HWR.beamline.diffractometer.wait_status_ready()
            except Exception as ex:
                logging.getLogger("MX3.HWR").exception("")
                msg = "Error while centring, please try again"
                self._click_in_progress = False
                raise RuntimeError(msg) from ex
        else:
            self.centring_reset_click_count()
            self._ho.cancel_centring()
            self._ho.start_manual_centring(
                HWR.beamline.config.click_centring_num_clicks
            )

        self._click_in_progress = False
        return {"clicksLeft": self.centring_clicks_left()}

    def accept_centring(self):
        self._ho.accept_centring()
        return {}

    def move_to_beam(self, x: float, y: float):
        self._ho.move_to_beam(x, y)
        return {}

    def start_auto_centring(self):
        """Start automatic centring procedure."""
        if not self._ho.current_centring_procedure:
            msg = "Starting automatic centring"
            logging.getLogger("user_level_log").info(msg)
            self._ho.start_auto_centring()
        else:
            msg = "Could not starting automatic centring, already centring."
            logging.getLogger("user_level_log").info(msg)

    def start_click_centring(self) -> dict:
        """Start click centring procedure.

        Returns:
            clicksLeft
        """
        if HWR.beamline.diffractometer.is_ready():
            self.centring_reset_click_count()
            self._error = False
            if self._ho.current_centring_procedure:
                logging.getLogger("user_level_log").info(
                    "Aborting current centring ..."
                )
                self._ho.reject_centring()
                self._click_in_progress = False
            nb_clicks = HWR.beamline.config.click_centring_num_clicks
            msg = f"Centring using {nb_clicks}-click centring"
            logging.getLogger("user_level_log").info(msg)

            self._ho.start_manual_centring(nb_clicks)
        else:
            logging.getLogger("user_level_log").warning(
                "Diffractometer is busy, cannot start centering"
            )
            msg = "Diffractometer is busy, cannot start centering"
            raise RuntimeError(msg)

        return {"clicksLeft": self.centring_clicks_left()}

    def set_centring_method(self, method: int) -> dict:
        if method == CENTRING_METHOD.LOOP:
            msg = "Using automatic loop centring when mounting samples"
            HWR.beamline.queue_manager.centring_method = CENTRING_METHOD.LOOP
        else:
            msg = "Using click centring when mounting samples"
            HWR.beamline.queue_manager.centring_method = CENTRING_METHOD.MANUAL

        logging.getLogger("user_level_log").info(msg)

        return {}



import logging
from pathlib import Path

import werkzeug
from flask import Flask, request
from flask_login import current_user
from flask_socketio import SocketIO
from werkzeug.exceptions import HTTPException
from werkzeug.middleware.proxy_fix import ProxyFix

from mxcubeweb.core.server.csp import init_csp
from mxcubeweb.core.server.db import init_database
from mxcubeweb.core.server.limiter import init_rate_limiter
from mxcubeweb.core.server.security import init_flask_security
from mxcubeweb.core.util import networkutils


def http_exception_handler(e):
    err_msg = f"Error while calling {request.path}"

    if e.code == 404:
        logging.getLogger("server_access").info(
            f"URL {request.path} not found on server"
        )
    else:
        logging.getLogger("server_access").exception(err_msg)

    return "Not found", 404


class Server:
    def __init__(  # noqa: PLR0913
        self,
        flask,
        flask_socketio,
        db_session,
        user_datastore,
        security,
        limiter=None,
    ):
        self.flask = flask
        self.flask_socketio = flask_socketio
        self.db_session = db_session
        self.user_datastore = user_datastore
        self.limiter = limiter
        self.security = security

        self.restrict = networkutils.auth_required
        self.require_control = networkutils.require_control
        self.ws_restrict = networkutils.ws_valid_login_only
        self.route = staticmethod(self.flask.route)

    def emit(self, *args, **kwargs):
        from flask import has_request_context

        if has_request_context() and current_user.is_authenticated:
            logging.getLogger("server_access").debug(
                f"{current_user.username} websocket emit: {args} {kwargs}"
            )
        self.flask_socketio.emit(*args, **kwargs)

    def run(self, cfg):
        if cfg.flask.CERT == "SIGNED" and cfg.flask.CERT_PEM and cfg.flask.CERT_KEY:
            ssl_context = werkzeug.serving.load_ssl_context(
                cfg.flask.CERT_PEM, cfg.flask.CERT_KEY
            )
        elif cfg.flask.CERT == "ADHOC":
            cert_dir = Path(cfg.flask.USER_DB_PATH).parent
            ssl_context = werkzeug.serving.load_ssl_context(
                *werkzeug.serving.make_ssl_devcert(str(cert_dir))
            )
        else:
            ssl_context = None

        if ssl_context:
            self.flask_socketio.run(
                self.flask,
                host=cfg.flask.HOST,
                port=cfg.flask.PORT,
                ssl_context=ssl_context,
            )
        else:
            self.flask_socketio.run(
                self.flask,
                host=cfg.flask.HOST,
                port=cfg.flask.PORT,
            )

    def close(self):
        try:
            if self.db_session:
                self.db_session.close()
        except Exception:
            logging.getLogger("MX3.HWR").exception("Error while closing DB session")


def create_server(cfg, cmdline_options):
    template_dir = str(Path(__file__).parent / "templates")

    flask = Flask(
        __name__,
        static_folder=cmdline_options.static_folder,
        static_url_path="",
        template_folder=template_dir,
    )

    flask.config.from_object(cfg.flask)
    flask.wsgi_app = ProxyFix(flask.wsgi_app)
    flask.register_error_handler(HTTPException, http_exception_handler)

    init_csp(flask, cfg)
    socketio = SocketIO(
        flask, manage_session=False, cors_allowed_origins=cfg.flask.ALLOWED_CORS_ORIGINS
    )
    limiter = init_rate_limiter(flask)

    db_session, user_datastore = init_database(cfg)
    security = init_flask_security(flask, user_datastore)

    return Server(
        flask=flask,
        flask_socketio=socketio,
        db_session=db_session,
        user_datastore=user_datastore,
        limiter=limiter,
        security=security,
    )



from typing import ClassVar

from mxcubecore.HardwareObjects import GenericDiffractometer, MiniDiff
from mxcubecore.HardwareObjects.abstract import AbstractDiffractometer

from mxcubeweb.core.adapter.adapter_base import AdapterBase
from mxcubeweb.core.models.adaptermodels import ChipCalibrationInputModel
from mxcubeweb.core.models.configmodels import ResourceHandlerConfigModel

resource_handler_config = ResourceHandlerConfigModel(
    commands=[
        "set_chip_layout",
        "set_phase",
        "use_position_for_calibration",
        "ir_auto_focus",
    ],
    attributes=["data", "get_value", "head_configuration"],
)


class DiffractometerAdapter(AdapterBase):
    SUPPORTED_TYPES: ClassVar[list[object]] = [
        MiniDiff.MiniDiff,
        GenericDiffractometer.GenericDiffractometer,
        AbstractDiffractometer.AbstractDiffractometer,
    ]

    def __init__(  # noqa: D417
        self, ho, role, app
    ):
        """Initialize.

        Args:
            ho (object): Hardware object.
        """
        super().__init__(ho, role, app, resource_handler_config)
        ho.connect("stateChanged", self._state_change)
        ho.connect("valueChanged", self._state_change)
        ho.connect("phaseChanged", self._diffractometer_phase_changed)

    def _diffractometer_phase_changed(self, phase):
        self.app.server.emit(
            "diff_phase_changed",
            {"msg": "Diffractometer phase changed", "phase": phase},
            namespace="/hwr",
        )

    def _state_change(self, *args, **kwargs):
        self.state_change(*args, **kwargs)

    def get_value(self) -> dict:
        return {
            "currentPhase": self._ho.get_phase().name,
            "phaseList": self._ho.get_phase_list(),
        }

    def stop(self):
        pass

    def state(self):
        return "READY" if self._ho.is_ready() else "BUSY"

    def head_configuration(self) -> dict:
        data = self._ho.get_chip_configuration()
        return data.dict() if data else {}

    def set_chip_layout(
        self,
        layout_name: str,
    ) -> bool:
        self._ho.set_chip_layout(layout_name)
        return True

    def set_phase(self, phase: str) -> bool:
        self._ho.set_phase(phase)
        return True

    def use_position_for_calibration(self, data: ChipCalibrationInputModel) -> bool:
        return self._ho.use_position_for_calibration(data.dict())

    def ir_auto_focus(self) -> bool:
        return self._ho.ir_auto_focus()



from mxcubeweb.core.components.user.database import UserDatastore, init_db
from mxcubeweb.core.models.usermodels import Message, Role, User


def init_database(cfg):
    db_session = init_db(cfg.flask.USER_DB_PATH)

    user_datastore = UserDatastore(
        db_session,
        User,
        Role,
        message_model=Message,
    )

    return db_session, user_datastore



import logging
from typing import ClassVar

from mxcubecore.HardwareObjects.abstract import AbstractBeam

from mxcubeweb.core.adapter.adapter_base import (
    ActuatorAdapterBase,
    default_resource_handler_config,
)
from mxcubeweb.core.models.adaptermodels import (
    HOBeamModel,
    HOBeamValueModel,
)


class BeamAdapter(ActuatorAdapterBase):
    SUPPORTED_TYPES: ClassVar[list[object]] = [AbstractBeam.AbstractBeam]

    def __init__(self, ho, role, app):
        super().__init__(ho, role, app, default_resource_handler_config)

        for sig in [
            "beamPosChanged",
            "beamInfoChanged",
            "valueChanged",
            "stateChanged",
        ]:
            ho.connect(ho, sig, self._beam_changed)

    def _beam_changed(self, *args, **kwargs):  # noqa: ARG002
        """Callback method for beam position and info changes."""
        self.app.server.emit(
            "beam_changed", {"data": self.get_value().dict()["value"]}, namespace="/hwr"
        )

    def limits(self):
        return -1, -1

    def _get_aperture(self) -> tuple:
        """Return list of apertures and the one currently used.

        :return: Tuple, (list of apertures, current aperture)
        :rtype: tuple
        """
        beam_ho = self._ho

        aperture_list = beam_ho.get_available_size()["values"]
        current_aperture = beam_ho.get_value()[-1]

        return aperture_list, current_aperture

    def set_value(self, value: int) -> HOBeamModel:
        self._ho.set_value(value)
        msg = f"Changing beams size to: {value}"
        logging.getLogger("MX3.HWR").info(msg)
        return self.get_value()

    def get_value(self) -> HOBeamValueModel:
        beam_info_dict = {
            "position": [],
            "shape": "",
            "size_x": 0,
            "size_y": 0,
        }
        sx, sy, shape, _label = self._ho.get_value()

        beam_info_dict.update(
            {
                "position": self._ho.get_beam_position_on_screen(),
                "size_x": sx,
                "size_y": sy,
                "shape": shape.value,
            }
        )

        aperture_list, current_aperture = self._get_aperture()

        beam_info_dict.update(
            {
                "apertureList": aperture_list,
                "currentAperture": current_aperture,
            }
        )

        return HOBeamModel(value={**beam_info_dict})

    def data(self) -> HOBeamModel:
        return HOBeamModel(**self._dict_repr())



import logging

import flask_security
from flask import redirect, request


def flask_security_custom_unauthn_handler(*args, **kwargs):  # noqa: ARG001
    logging.getLogger("server_access").exception(
        f"Authentication failed on {request.path}"
    )

    return redirect("/login")


def init_flask_security(app, user_datastore):
    app.config.update(
        SECURITY_LOGIN_URL=None,
        SECURITY_REDIRECT_BEHAVIOR="spa",
    )

    security = flask_security.Security()
    security.init_app(
        app,
        user_datastore,
        register_blueprint=False,
    )

    security.unauthz_handler(flask_security_custom_unauthn_handler)
    security.unauthn_handler(flask_security_custom_unauthn_handler)

    return security



import email.utils
import functools
import logging
import os
import smtplib
import socket
import time
from email.mime.text import MIMEText
from email.utils import make_msgid

import flask
import flask_security
import flask_socketio
from flask_login import current_user
from mxcubecore import HardwareRepository as HWR


def auth_required(fun):
    return flask_security.auth_required("session", within=-1, grace=None)(fun)


def RateLimited(maxPerSecond):
    minInterval = 1.0 / float(maxPerSecond)
    lastTimeCalled = {}

    def decorate(func):
        def rateLimitedFunction(*args, **kargs):
            key = args[0].get("Signal") if type(args[0]) is dict else args[0]
            elapsed = time.time() - lastTimeCalled.get(key, 0)
            leftToWait = minInterval - elapsed
            if leftToWait > 0:
                # ignore update
                return None
            ret = func(*args, **kargs)
            lastTimeCalled.update({key: time.time()})
            return ret

        return rateLimitedFunction

    return decorate


def remote_addr():
    hdr = flask.request.headers.get("x-forwarded-for", flask.request.remote_addr)

    return str(hdr).split(",")[-1]


def is_local_network(ip: str, local_domains: list):
    """Determine whether a given IP address belongs to the local network.

    The function compares the first two octets of the given IP address with the
    local host's IP address. It also checks if the reverse-resolved hostname of
    the IP ends with any of the specified local domains.

    Args:
        ip (str): The IP address to check.
        local_domains (list): A list of domain suffixes that are considered local
            (e.g. ['beamline1.site.eu', 'control.site.eu']).

    Returns:
        bool: True if the IP is in the same network range as the local host
              (i.e. mxcube server) or its hostname ends with any of the specified
              local domains; False otherwise.
    """
    localhost = socket.gethostbyname_ex(socket.gethostname())[2][0]
    localhost_range = ".".join(localhost.split(".")[0:2])
    private_address = ".".join(ip.split(".")[0:2])
    try:
        private_hostname = socket.gethostbyaddr(ip)[0]
    except (socket.herror, socket.gaierror):
        logging.getLogger("MX3.HWR").warning("Failed to resolve the client IP: %s" % ip)
        private_hostname = ""

    return private_address == localhost_range or any(
        private_hostname and private_hostname.endswith(domain)
        for domain in local_domains
    )


def is_local_host(local_domains: list):
    """Determine whether the remote client making the request is a "local host".

    A client is considered a "local host" if it runs on the same host as the
    mxcube-server, or if it belongs to the local network. The local network can
    also be identified by providing a list of domain suffixes considered local.

    Args:
        local_domains (list): A list of domain suffixes that are considered local
            (e.g. ['beamline1.site.eu', 'control.site.eu']).

    Returns:
        bool: True if the client's host is considered a local host; False otherwise.
    """
    try:
        localhost_list = socket.gethostbyname_ex(socket.gethostname())[2]
    except Exception:
        localhost_list = []

    localhost_list.append("127.0.0.1")

    remote_address = remote_addr()

    # Remote address is sometimes None for instance when using the test
    # client, no real connection is made, assume that we are local host
    if remote_address in [None, "None", ""]:
        remote_address = "127.0.0.1"

    return remote_address in localhost_list or is_local_network(
        remote_address, local_domains
    )


def valid_login_only(f):
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        if not current_user.is_authenticated:
            return flask.Response(status=404)
        return f(*args, **kwargs)

    return wrapped


def require_control(f):
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        if current_user.is_authenticated and not current_user.in_control:
            return flask.Response(status=401)
        return f(*args, **kwargs)

    return wrapped


def ws_valid_login_only(f):
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        if not current_user.is_authenticated:
            flask_socketio.disconnect()
            return None
        return f(*args, **kwargs)

    return wrapped


def send_mail(_from, to, subject, content):
    smtp = smtplib.SMTP("smtp", smtplib.SMTP_PORT)
    date = email.utils.formatdate(localtime=True)

    msg = MIMEText(content)
    msg["Subject"] = subject
    msg["From"] = _from
    msg["To"] = to
    msg["Date"] = date
    msg["Message-ID"] = make_msgid()

    email_msg = msg.as_string()

    try:
        error_dict = smtp.sendmail(_from, to.split(","), email_msg)

        if error_dict:
            msg = "Could not send mail to %s, content %s, error was: %s"
            msg = msg % (to, content, str(error_dict))
            logging.getLogger().error(msg)
        else:
            msg = "Feedback sent to %s, msg: \n %s" % (to, content)
            logging.getLogger("MX3.HWR").info(msg)

    except smtplib.SMTPException as e:
        msg = "Could not send mail to %s, content %s, error was: %s"
        logging.getLogger().error(msg % (to, content, str(e)))
    finally:
        smtp.quit()


def send_feedback(sender_data):
    bl_name = HWR.beamline.session.beamline_name
    local_user = sender_data.get("LOGGED_IN_USER", "")

    if not bl_name:
        try:
            bl_name = os.environ["BEAMLINENAME"].lower()
        except KeyError:
            bl_name = "unknown-beamline"

    if not local_user:
        try:
            local_user = os.environ["USER"].lower()
        except KeyError:
            local_user = "unknown_user"

    _from = HWR.beamline.session.get_property("from_email", "")

    if not _from:
        _from = "%s@%s" % (
            local_user,
            HWR.beamline.session.get_property("email_extension", ""),
        )

    # Sender information provided by user
    _sender = sender_data.get("sender", "")
    to = HWR.beamline.session.get_property("feedback_email", "") + ",%s" % _sender
    subject = "[MX3 FEEDBACK] %s (%s) on %s" % (
        local_user,
        _sender,
        bl_name,
    )
    content = sender_data.get("content", "")

    send_mail(_from, to, subject, content)



import datetime
from enum import Enum
from pathlib import Path
from typing import (
    Literal,
)

from pydantic import BaseModel, Field, validator
from pydantic_settings import BaseSettings, SettingsConfigDict


class RuntimeOptions(BaseSettings):
    hwr_directory: str
    static_folder: str | None = Path.cwd() / "ui" / "build"
    log_file: str = ""
    log_level: str = ""
    enabled_logger_list: list[str] = [
        "exception_logger",
        "hwr_logger",
        "server_logger",
        "user_logger",
        "queue_logger",
        "ui_logger",
        "csp_logger",
        "server_access_logger",
    ]
    allow_remote: bool = False
    ra_timeout: bool = False
    yaml_export_directory: str | None = None

    # Pydantic-settings config
    model_config = SettingsConfigDict(
        env_prefix="MXCUBE_",  # env vars like MXCUBE_HWR_DIRECTORY
        case_sensitive=False,
        env_file=".env",  # automatically load .env file if present
        env_file_encoding="utf-8",
    )


class FlaskConfigModel(BaseModel):
    SECRET_KEY: str = Field(
        b"o`\xb5\xa5\xc2\x8c\xb2\x8c-?\xe0,/i#c",
        description="Flask secret key",
    )
    DEBUG: bool = Field(False, description="")
    ALLOWED_CORS_ORIGINS: list[str] = Field([], description="")
    SECURITY_PASSWORD_SALT: str = Field("ASALT", description="")
    SECURITY_TRACKABLE: bool = Field(True, description="")
    USER_DB_PATH: str = Field(
        str(Path.home() / ".local" / "share" / "mxcube" / "mxcube-user.db"),
        description="",
    )
    HOST: str = Field("127.0.0.1", description="Host address to bind to")
    PORT: int = Field(8081, description="Port to bind to")
    PERMANENT_SESSION_LIFETIME: datetime.timedelta
    CERT_KEY: str = Field("", description="Full path to signed certificate key file")
    CERT_PEM: str = Field("", description="Full path to signed certificate pem file")

    # SIGNED for signed certificate on file
    # ADHOC for flask to generate a certificate,
    # NONE for no SSL
    CERT: str = Field(
        "NONE",
        description="One of the strings ['SIGNED', 'ADHOC', NONE]",
    )
    # Rate limiter configuration
    RATE_LIMITER_ENABLED: bool = True
    RATELIMIT_DEFAULT: str = "150000 per day;6000 per hour"
    RATELIMIT_STORAGE_URI: str = "memory://"
    RATELIMIT_HEADERS_ENABLED: bool = True

    CSP_ENABLED: bool = True
    CSP_POLICY: dict[str, list[str]] = {}
    CSP_REPORT_ONLY: bool = Field(
        False,
        description="Set to True to enable report-only mode (won't block content)",
    )
    CSP_REPORT_URI: str = Field(
        "", description="URI for CSP violation reports (empty to disable reporting)"
    )


class SSOConfigModel(BaseModel):
    USE_SSO: bool = Field(False, description="Set to True to use SSO")
    ISSUER: str = Field("", description="OpenIDConnect / OAuth Issuer URI")
    LOGOUT_URI: str = Field("", description="OpenIDConnect / OAuth logout URI")
    CLIENT_SECRET: str = Field("", description="OpenIDConnect / OAuth client secret")
    CLIENT_ID: str = Field("", description="OpenIDConnect / OAuth  client id")
    META_DATA_URI: str = Field(
        "", description="OpenIDConnect / OAuth  .well-known configuration"
    )
    SCOPE: str = Field(
        "openid email profile", description="OpenIDConnect / OAuth scope"
    )
    CODE_CHALLANGE_METHOD: str = Field(
        "S256", description="OpenIDConnect / OAuth Challange"
    )


class UIComponentModel(BaseModel):
    label: str
    attribute: str
    role: str | None = None
    step: float | None = None
    precision: int | None = None
    suffix: str | None = None
    description: str | None = None
    # Set internally not to be set through configuration
    value_type: str | None = None
    object_type: str | None = None
    format: str | None = None
    invert_color_semantics: bool | None = None


class _UICameraConfigModel(BaseModel):
    label: str
    url: str
    format: str | None = None
    description: str | None = None
    width: int | None = None
    height: int | None = None


class _UISampleViewVideoControlsModel(BaseModel):
    id: str
    show: bool


class _UISampleViewVideoGridSettingsModel(BaseModel):
    id: Literal["draw_grid"]
    show: bool
    show_vspace: bool = False
    show_hspace: bool = False


class UIPropertiesModel(BaseModel):
    id: str
    components: list[UIComponentModel]


class UICameraConfigModel(UIPropertiesModel):
    components: list[_UICameraConfigModel]


class UISampleViewMotorsModel(UIPropertiesModel):
    id: str
    components: list[UIComponentModel]


class _UISampleListViewModesComponentModel(BaseModel):
    id: Literal["table_view", "graphical_view"]
    show: bool = True


class UISampleListViewModesModel(BaseModel):
    id: Literal["sample_list_view_modes"] = "sample_list_view_modes"
    components: list[_UISampleListViewModesComponentModel] = Field(
        default_factory=lambda: [
            _UISampleListViewModesComponentModel(id="table_view", show=True),
            _UISampleListViewModesComponentModel(id="graphical_view", show=True),
        ],
        description="List of components for sample list view modes",
    )

    @validator("components")
    @classmethod
    def check_at_least_one_component_shown(
        cls, components: list[_UISampleListViewModesComponentModel]
    ) -> list[_UISampleListViewModesComponentModel]:
        """Validate that at least one component in the list has 'show' set to True."""
        if not any(component.show for component in components):
            msg = "At least one component must have 'show' set to True."
            raise ValueError(msg)
        return components


class UISampleViewVideoControlsModel(UIPropertiesModel):
    # It is important to keep the Union elements in that order; from the more specific to the more general.
    components: list[
        _UISampleViewVideoGridSettingsModel | _UISampleViewVideoControlsModel
    ]


class UiSessionPickerTabs(BaseModel):
    active: bool = True
    scheduled: bool = True
    other_beamlines: bool = True
    all_sessions: bool = False


class UISessionPickerModel(BaseModel):
    id: Literal["session_picker"] = "session_picker"
    tabs: UiSessionPickerTabs = UiSessionPickerTabs()


class UIPropertiesListModel(BaseModel):
    sample_view: UIPropertiesModel | None = None
    beamline_setup: UIPropertiesModel
    camera_setup: UICameraConfigModel | None = None
    sample_view_motors: UISampleViewMotorsModel
    sample_list_view_modes: UISampleListViewModesModel = UISampleListViewModesModel()
    sample_view_video_controls: UISampleViewVideoControlsModel | None = None
    session_picker: UISessionPickerModel = UISessionPickerModel()


class UserManagerUserConfigModel(BaseModel):
    username: str = Field("", description="username")
    role: str = Field("staff", description="Role to give user")


class UserManagerConfigModel(BaseModel):
    class_name: str = Field(
        "UserManager", description="UserManager class", alias="class"
    )
    inhouse_is_staff: bool = Field(
        True,
        description="Treat users defined as inhouse in session.xml as staff",
    )
    users: list[UserManagerUserConfigModel]


class ModeEnum(str, Enum):
    SSX_INJECTOR = "SSX-INJECTOR"
    SSX_CHIP = "SSX-CHIP"
    OSC = "OSC"


class MXCUBEAppConfigModel(BaseModel):
    VIDEO_FORMAT: str = Field("MPEG1", description="Video format MPEG1 or MJPEG")

    # URL from which the client retrieves the video stream (often different from
    # local host when running behind proxy)
    VIDEO_STREAM_URL: str = Field(
        "",
        description="Video stream URL, URL used by client to get video stream",
    )

    # Port from which the video_stream process (https://github.com/mxcube/video-streamer)
    # streams video. The process runs in separate process (on localhost)
    VIDEO_STREAM_PORT: int = Field(8000, description="Video stream PORT")

    # Should we call the `start_streaming()` method of Camera HWO
    MXCUBE_STARTS_VIDEO_STREAM: bool = Field(
        True,
        description="Invoke 'start OAV stream' method on the Camera hardware object",
    )

    USE_GET_SAMPLES_FROM_SC: bool = Field(
        True,
        description=(
            "True to activate or be able to get samples from the sample changer, false otherwise"
        ),
    )

    AUTOSYNC_LIMS: bool = Field(
        False, description="True to synchronize samples with LIMS on proposal selection"
    )

    mode: ModeEnum = Field(
        ModeEnum.OSC, description="MXCuBE mode OSC, SSX-CHIP or SSX-INJECTOR"
    )
    LOCAL_DOMAINS: list[str] = Field(
        [],
        description="If the connected client's hostname ends with one of these domains"
        ", it will be considered 'local'",
    )
    SESSION_REFRESH_INTERVAL: int = Field(
        9000,
        description="Session refresh interval in milliseconds",
    )
    usermanager: UserManagerConfigModel
    ui_properties: dict[str, UIPropertiesModel] = {}


class BraggyConfigModel(BaseModel):
    BRAGGY_URL: str = Field("", description="Base URL for braggy server")
    USE_BRAGGY: bool = Field(
        False, description="Set to True to use braggy for displaying images"
    )


class AppConfigModel(BaseModel):
    server: FlaskConfigModel
    mxcube: MXCUBEAppConfigModel
    sso: SSOConfigModel | None = None
    braggy: BraggyConfigModel | None = None


class ResourceHandlerConfigModel(BaseModel):
    """Configuration modle for resource handler.

    Used to define  which adapter properties and methods that are
    exported over HTTP. An endpoint for each method and/or property
    is created by the AdapterResourceHandler and attached to the server.

    The exports list defines the methods and properties exported,
    the HTTP verb to use and the decorators to apply to the view
    function,

    The format is::

        [
            {"attr": "get_value", "method": "PUT", "decorators": []},
        ]

    Where "attr" is the method or property of the adapter, "method"
    is the http verb i.e GET, PUT, POST. and decoratoes are a list
    decroator functions to apply to the resulting view function.

    There are two structures that can be used for convenience, that
    use defualt vlaues for "method" and "decorators". These are
    commands and attributes.

    Commands is list of functions/methods to export.
    A dictionary like the one above::

        ({"attr": "get_value", "method": "PUT", "decorators": []})

    Will be generated from the commands list. With the values "method" and
    decorators set to defualt values, PUT and [restrict, require_control]

    Example::

        commands = ["set_value", "get_value"]

    Will export the methods as .../set_value with HTTP verb PUT

    In the same way:

    Example::

        attributes = ["data"]

    Will export the property as .../data with HTTP verb GET
    """

    url_prefix: str = Field("", description="URL prefix")
    name: str | None = Field(
        None,
        description=(
            "Name of the resource handler, if not given the class name (lower case) ",
            "is used",
        ),
    )
    exports: list[dict[str, str | list]] = Field(
        [],
        description=(
            "List of dictionaires specifying each of the exported attributes or"
            " methods, HTTP method to use and decorators to apply "
        ),
    )
    commands: list[str] = Field(
        [], description="List of exported methods, defualted to HTTP PUT"
    )
    attributes: list[str] = Field(
        [], description="List of exported properties, defaulted to HTTP GET"
    )



from mxcubeweb.core.server.resource_handler import ResourceHandlerFactory
from mxcubeweb.routes.csp_report import init_route as init_csp_route
from mxcubeweb.routes.harvester import init_route as init_harvester_route
from mxcubeweb.routes.login import init_route as init_login_route
from mxcubeweb.routes.main import init_route as init_main_route
from mxcubeweb.routes.queue import init_route as init_queue_route
from mxcubeweb.routes.ra import init_route as init_ra_route
from mxcubeweb.routes.workflow import init_route as init_workflow_route


def register_routes(server, mxcube_app, cfg):  # noqa: ARG001
    url_root_prefix = "/mxcube/api/v0.1"  # cfg.flask.API_PREFIX

    _register_route(server, init_csp_route, mxcube_app, f"{url_root_prefix}/csp")
    _register_route(server, init_login_route, mxcube_app, f"{url_root_prefix}/login")
    _register_route(server, init_main_route, mxcube_app, f"{url_root_prefix}")
    _register_route(server, init_queue_route, mxcube_app, f"{url_root_prefix}/queue")
    _register_route(server, init_ra_route, mxcube_app, f"{url_root_prefix}/ra")
    _register_route(
        server, init_workflow_route, mxcube_app, f"{url_root_prefix}/workflow"
    )
    _register_route(
        server, init_harvester_route, mxcube_app, f"{url_root_prefix}/harvester"
    )

    ResourceHandlerFactory.register_with_server(server.flask)


def _register_route(server, init_blueprint_fn, app, url_prefix):
    bp = init_blueprint_fn(app, server, url_prefix)
    server.flask.register_blueprint(bp)

    for key, function in server.flask.view_functions.items():
        if key.startswith(bp.name) and not hasattr(function, "tags"):
            function.tags = [bp.name.title().replace("_", " ")]

    return bp
